@@ -59,6 +59,8 @@ GST_STATIC_PAD_TEMPLATE ("src%d",
 GST_DEBUG_CATEGORY_STATIC (gst_rtp_bin_debug);
 #define GST_CAT_DEFAULT gst_rtp_bin_debug
 
+#define UDP_DEFAULT_CLOSEFD            TRUE
+
 /* props */
 enum
 {
@@ -74,7 +76,8 @@ enum
   ARG_PT_MAP,
   ARG_PT_CAPS,
   ARG_BYPASS_UDP,
-  ARG_QUEUE_DELAY
+  ARG_QUEUE_DELAY,
+  ARG_CLOSE_FD
 };
 
 
@@ -105,7 +108,8 @@ static gboolean gst_rtp_bin_setup_recv_elements (GstRTPBin *rtp_bin);
 static GstStateChangeReturn
 gst_rtp_bin_change_state (GstElement *element, GstStateChange transition);
 
-static void gst_rtp_bin_set_destinations (GstRTPBin *rtp_bin, const gchar *destination);
+static void gst_rtp_bin_set_destinations (GstRTPBin *rtp_bin,
+    const gchar *destination);
 
 static void gst_rtp_bin_set_property (GObject * object, guint prop_id, 
         const GValue * value, GParamSpec * spec);
@@ -226,6 +230,11 @@ gst_rtp_bin_class_init (GstRTPBinClass * klass)
 	    g_param_spec_uint ("queue-delay", "Queue Delay",
 			       "Amount of ms to queue/buffer, or zero to disable", 0, G_MAXUINT, 0, G_PARAM_READWRITE));
 
+  g_object_class_install_property (gobject_klass, ARG_CLOSE_FD,
+      g_param_spec_boolean ("closefd", "Close sockfd",
+          "Close sockfd if passed as property on state change",
+          UDP_DEFAULT_CLOSEFD, G_PARAM_READWRITE));
+
     gst_element_class_add_pad_template (gstelement_klass,
             gst_static_pad_template_get (&rtp_bin_sink_template));
     gst_element_class_add_pad_template (gstelement_klass,
@@ -277,6 +286,8 @@ gst_rtp_bin_init (GstRTPBin * rtp_bin)
     rtp_bin->default_pt_map = gst_rtp_bin_build_pt_map (default_pt_table);
 
     rtp_bin->pt_map = NULL;
+
+    rtp_bin->closefd = UDP_DEFAULT_CLOSEFD;
 }
 
 static void
@@ -331,7 +342,8 @@ static GHashTable *gst_rtp_bin_build_pt_map (const gchar *caps_string)
   for (i = 0; i < gst_caps_get_size (caps); i++)
   {
     GstStructure *structure = gst_caps_get_structure (caps, i);
-    GST_DEBUG ("Adding struct %p %s", structure, gst_structure_to_string (structure));
+    GST_DEBUG ("Adding struct %p %s", structure,
+        gst_structure_to_string (structure));
     GstCaps *new_caps = gst_caps_new_full (gst_structure_copy (structure), NULL);
     GST_DEBUG ("Adding caps %p %s", new_caps, gst_caps_to_string (new_caps));
     gint pt;
@@ -379,6 +391,8 @@ gst_rtp_bin_request_new_pad (GstElement * element, GstPadTemplate * templ,
                 goto error;
             }
 
+            gst_element_sync_state_with_parent (rtp_bin->rtpsend);
+
             newpad = gst_rtp_bin_setup_rtpsend_pads(rtp_bin);
             if (!newpad)
             {
@@ -386,6 +400,21 @@ gst_rtp_bin_request_new_pad (GstElement * element, GstPadTemplate * templ,
                 goto error;
             }
 
+            /* Setup send elements if this pad was added after changing
+               to the READY state. */
+            if (GST_STATE(rtp_bin) >= GST_STATE_READY)
+            {
+                if (!rtp_bin->bypass_udp)
+                {
+                    if (!gst_rtp_bin_setup_send_elements(rtp_bin)) {
+                        goto error;
+                    }
+                    /* Set children to the same state as parent so they
+                       start running if rtpbin is already in PLAYING state. */
+                    gst_element_sync_state_with_parent(rtp_bin->rtpsink);
+                }
+            }
+
             break;
 
         case GST_PAD_SRC:          /* Create rtprecv and its ghost pad. */
@@ -401,12 +430,18 @@ gst_rtp_bin_request_new_pad (GstElement * element, GstPadTemplate * templ,
                     == NULL) {
                 goto error;
             }
-            g_object_set (rtp_bin->jbuf, "queue_delay", rtp_bin->queue_delay,
+            g_object_set (rtp_bin->jbuf, "latency", rtp_bin->queue_delay,
+                    NULL);
+            g_object_set (rtp_bin->jbuf, "drop-on-latency", TRUE,
                     NULL);
 
             gst_bin_add_many (GST_BIN (rtp_bin), rtp_bin->rtprecv,
                     rtp_bin->jbuf, NULL);
 
+            gst_element_sync_state_with_parent (rtp_bin->rtprecv);
+            gst_element_sync_state_with_parent (rtp_bin->jbuf);
+
+
             if (!gst_element_link_pads (rtp_bin->rtprecv, "datasrc",
                         rtp_bin->jbuf, "sink")) {
                 GST_ERROR ("Could not link rtprecv:datasrc to jitterbuffer:sink");
@@ -421,6 +456,22 @@ gst_rtp_bin_request_new_pad (GstElement * element, GstPadTemplate * templ,
                 goto error;
             }
 
+            /* Setup recv elements if this pad was added after changing
+               to the READY state. */
+            if (GST_STATE(rtp_bin) >= GST_STATE_READY)
+            {
+                if (!rtp_bin->bypass_udp)
+                {
+                    if (!gst_rtp_bin_setup_recv_elements(rtp_bin)) {
+                        goto error;
+                    }
+                    /* Set children to the same state as parent so they
+                       start running if rtpbin is already in PLAYING state. */
+                    gst_element_sync_state_with_parent(rtp_bin->rtpsrc);
+                    gst_element_sync_state_with_parent(rtp_bin->jbuf);
+                }
+            }
+
             break;
 
         default:
@@ -518,7 +569,7 @@ gst_rtp_bin_setup_elements (GstRTPBin * rtp_bin)
             GST_DEBUG ("Running gst_rtp_bin_setup_recv_elements");
             if (!gst_rtp_bin_setup_recv_elements (rtp_bin))
             {
-                return FALSE;
+              goto error;
             }
         }
     }
@@ -531,10 +582,12 @@ gst_rtp_bin_setup_elements (GstRTPBin * rtp_bin)
         if (rtp_bin->rtpsend == NULL) {
             rtp_bin->rtpsend = gst_rtp_bin_create_rtpsend (rtp_bin);
             gst_bin_add (GST_BIN(rtp_bin), rtp_bin->rtpsend);
+            gst_element_sync_state_with_parent (rtp_bin->rtpsend);
         }
         if (rtp_bin->rtprecv == NULL) {
             rtp_bin->rtprecv = gst_rtp_bin_create_rtprecv (rtp_bin);
             gst_bin_add (GST_BIN(rtp_bin), rtp_bin->rtprecv);
+            gst_element_sync_state_with_parent (rtp_bin->rtprecv);
         }
 
         /* if bypassing udp elements let's not create them */
@@ -543,7 +596,7 @@ gst_rtp_bin_setup_elements (GstRTPBin * rtp_bin)
             GST_DEBUG ("Running gst_rtp_bin_setup_rtcp_elements");
             if (!gst_rtp_bin_setup_rtcp_elements (rtp_bin))
             {
-                return FALSE;
+              goto error;
             }
             rtcp_src_pad = gst_element_get_pad (rtp_bin->rtpsend, "rtcpsrc");
         }
@@ -553,16 +606,20 @@ gst_rtp_bin_setup_elements (GstRTPBin * rtp_bin)
             /* Let's just ghost pad rtcpsink on rtpbin */
             GstPad *newpad1 = NULL;
             GstPad *newpad2 = NULL;
-            newpad1 = gst_ghost_pad_new ("rtcpsink",
-                    gst_element_get_pad (rtp_bin->rtprecv, "rtcpsink"));
+            GstPad *tmppad = NULL;
+
+            tmppad = gst_element_get_static_pad (rtp_bin->rtprecv, "rtcpsink");
+            newpad1 = gst_ghost_pad_new ("rtcpsink", tmppad );
+            gst_object_unref (tmppad);
 
             if (!gst_element_add_pad (GST_ELEMENT (rtp_bin), newpad1)) {
                 gst_object_unref (newpad1);
                 GST_DEBUG ("Could not setup rtcpsink ghost pad");
                 return FALSE;
             }
-            newpad2 = gst_ghost_pad_new ("rtcpsrc",
-                    gst_element_get_pad (rtp_bin->rtpsend, "rtcpsrc"));
+            tmppad = gst_element_get_static_pad (rtp_bin->rtpsend, "rtcpsrc");
+            newpad2 = gst_ghost_pad_new ("rtcpsrc", tmppad);
+            gst_object_unref (tmppad);
 
             if (!gst_element_add_pad (GST_ELEMENT (rtp_bin), newpad2)) {
                 gst_object_unref (newpad2);
@@ -575,10 +632,24 @@ gst_rtp_bin_setup_elements (GstRTPBin * rtp_bin)
 
 #ifdef HAVE_JRTP
     jrtpsession_setpads (rtp_bin->sess, rtp_src_pad, rtcp_src_pad);
+#else
+    if (rtp_src_pad)
+      gst_object_unref(rtp_src_pad);
+    if (rtcp_src_pad)
+      gst_object_unref(rtcp_src_pad);
 #endif
 
-    GST_DEBUG("Elements setup propertly");
+    GST_DEBUG("Elements setup properly");
     return TRUE;
+
+ error:
+
+    if (rtp_src_pad)
+      gst_object_unref(rtp_src_pad);
+    if (rtcp_src_pad)
+      gst_object_unref(rtcp_src_pad);
+
+    return FALSE;
 }
 
 static gboolean
@@ -602,12 +673,20 @@ gst_rtp_bin_setup_rtcp_elements (GstRTPBin * rtp_bin)
 
     /* If application provided a socket, use it */
     if (rtp_bin->rtcp_sockfd != -1) {
-      g_object_set (G_OBJECT(rtp_bin->rtcpsrc), "sockfd", rtp_bin->rtcp_sockfd, NULL);
-      g_object_set (G_OBJECT(rtp_bin->rtcpsink), "sockfd", rtp_bin->rtcp_sockfd, NULL);
+      g_object_set (G_OBJECT(rtp_bin->rtcpsrc), "sockfd",
+          rtp_bin->rtcp_sockfd, NULL);
+      g_object_set (G_OBJECT(rtp_bin->rtcpsrc), "closefd",
+          rtp_bin->closefd, NULL);
+
+      g_object_set (G_OBJECT(rtp_bin->rtcpsink), "sockfd",
+          rtp_bin->rtcp_sockfd, NULL);
+      g_object_set (G_OBJECT(rtp_bin->rtcpsink), "closefd",
+          rtp_bin->closefd, NULL);
     }
     g_object_set (G_OBJECT(rtp_bin->rtcpsink), "sync", FALSE, NULL);
 
-    gst_bin_add_many (GST_BIN(rtp_bin), rtp_bin->rtcpsink, rtp_bin->rtcpsrc, NULL);
+    gst_bin_add_many (GST_BIN(rtp_bin), rtp_bin->rtcpsink,
+        rtp_bin->rtcpsrc, NULL);
     /* Set the local RTCP port */
     g_object_set (G_OBJECT(rtp_bin->rtcpsrc), "port", rtp_bin->localport+1,
         NULL);
@@ -644,9 +723,14 @@ gst_rtp_bin_setup_send_elements (GstRTPBin * rtp_bin)
 
     /* If application provided a socket, use it */
     if (rtp_bin->rtp_sockfd != -1)
-      g_object_set (G_OBJECT(rtp_bin->rtpsink), "sockfd", rtp_bin->rtp_sockfd, NULL);
+    {
+      g_object_set (G_OBJECT(rtp_bin->rtpsink), "sockfd", rtp_bin->rtp_sockfd,
+          NULL);
+      g_object_set (G_OBJECT(rtp_bin->rtpsink), "closefd", rtp_bin->closefd,
+          NULL);
+    }
 
-    //g_object_set (G_OBJECT(rtp_bin->rtpsink), "sync", FALSE, NULL);
+    g_object_set (G_OBJECT(rtp_bin->rtpsink), "sync", FALSE, NULL);
 
     gst_bin_add (GST_BIN (rtp_bin), rtp_bin->rtpsink);
 
@@ -675,7 +759,12 @@ gst_rtp_bin_setup_recv_elements (GstRTPBin * rtp_bin)
 
     /* If application provided a socket, use it */
     if (rtp_bin->rtp_sockfd != -1)
-      g_object_set (G_OBJECT(rtp_bin->rtpsrc), "sockfd", rtp_bin->rtp_sockfd, NULL);
+    {
+      g_object_set (G_OBJECT(rtp_bin->rtpsrc), "sockfd", rtp_bin->rtp_sockfd,
+          NULL);
+      g_object_set (G_OBJECT(rtp_bin->rtpsrc), "closefd", rtp_bin->closefd,
+          NULL);
+    }
 
     /* Set our local port. */
     g_object_set (G_OBJECT(rtp_bin->rtpsrc), "port", rtp_bin->localport, NULL);
@@ -737,7 +826,8 @@ gst_rtp_bin_create_rtprecv (GstRTPBin * rtp_bin)
     else
     {
       GST_DEBUG_OBJECT (rtp_bin, "Using default pt-map table");
-      g_object_set (G_OBJECT(rtprecv), "pt_map", rtp_bin->default_pt_map, NULL);
+      g_object_set (G_OBJECT(rtprecv), "pt_map", rtp_bin->default_pt_map,
+          NULL);
     }
 
     return rtprecv;
@@ -747,12 +837,17 @@ static GstElement *
 gst_rtp_bin_create_jitterbuffer (GstRTPBin * rtp_bin)
 {
     GstElement * rtpjitterbuffer = NULL;
+    const gchar * jb_name;
 
-    rtpjitterbuffer = gst_element_factory_make ("rtpjitterbuffer",
+    jb_name = g_getenv ("RTPBIN_JB");
+    if (jb_name == NULL)
+      jb_name = "rtpjitterbuffer";
+
+    rtpjitterbuffer = gst_element_factory_make (jb_name,
             NULL);
     if (rtpjitterbuffer == NULL) {
-        GST_WARNING("Could not create rtpjitterbuffer element");
-        return NULL;
+      GST_WARNING("Could not create %s element", jb_name);
+      return NULL;
     }
 
     return rtpjitterbuffer;
@@ -762,8 +857,14 @@ static GstPad *
 gst_rtp_bin_setup_rtpsend_pads(GstRTPBin * rtp_bin)
 {
     GstPad *newpad = NULL;
-    newpad = gst_ghost_pad_new ("sink",
-            gst_element_get_pad (rtp_bin->rtpsend, "datasink"));
+    GstPad *internal_pad = NULL;
+
+    internal_pad = gst_element_get_pad (rtp_bin->rtpsend, "datasink");
+    newpad = gst_ghost_pad_new ("sink", internal_pad);
+
+    gst_pad_set_active (newpad, TRUE);
+
+    gst_object_unref (internal_pad);
 
     if (!gst_element_add_pad (GST_ELEMENT (rtp_bin), newpad)) {
         gst_object_unref (newpad);
@@ -775,8 +876,13 @@ gst_rtp_bin_setup_rtpsend_pads(GstRTPBin * rtp_bin)
         GST_DEBUG ("Using bypass_udp to create ghostpads for rtpsend");
         /* Let's just ghost pad rtpsrc on rtpbin */
         GstPad *bypass_pad = NULL;
-        bypass_pad = gst_ghost_pad_new ("rtpsrc",
-                gst_element_get_pad (rtp_bin->rtpsend, "rtpsrc"));
+        GstPad *tmppad = NULL;
+
+        tmppad = gst_element_get_static_pad (rtp_bin->rtpsend, "rtpsrc");
+        bypass_pad = gst_ghost_pad_new ("rtpsrc", tmppad);
+        gst_object_unref (tmppad);
+
+        gst_pad_set_active (bypass_pad, TRUE);
 
         if (!gst_element_add_pad (GST_ELEMENT (rtp_bin), bypass_pad)) {
             gst_object_unref (bypass_pad);
@@ -791,8 +897,13 @@ static GstPad *
 gst_rtp_bin_setup_rtprecv_pads(GstRTPBin * rtp_bin)
 {
     GstPad *newpad = NULL;
-    newpad = gst_ghost_pad_new ("src",
-            gst_element_get_pad (rtp_bin->jbuf, "src"));
+    GstPad *tmppad = NULL;
+
+    tmppad = gst_element_get_static_pad (rtp_bin->jbuf, "src");
+    newpad = gst_ghost_pad_new ("src",tmppad);
+    gst_object_unref (tmppad);
+
+    gst_pad_set_active (newpad, TRUE);
 
     if (!gst_element_add_pad (GST_ELEMENT (rtp_bin), newpad)) {
         gst_object_unref (newpad);
@@ -804,8 +915,13 @@ gst_rtp_bin_setup_rtprecv_pads(GstRTPBin * rtp_bin)
         GST_DEBUG ("Using bypass_udp to create ghostpads on rtprecv");
         /* Let's just ghost pad rtpsink on rtpbin */
         GstPad *newpad = NULL;
-        newpad = gst_ghost_pad_new ("rtpsink",
-                gst_element_get_pad (rtp_bin->rtprecv, "rtpsink"));
+        GstPad *tmppad = NULL;
+
+        tmppad = gst_element_get_static_pad (rtp_bin->rtprecv, "rtpsink");
+        newpad = gst_ghost_pad_new ("rtpsink",tmppad);
+        gst_object_unref (tmppad);
+
+        gst_pad_set_active (newpad, TRUE);
 
         if (!gst_element_add_pad (GST_ELEMENT (rtp_bin), newpad)) {
             gst_object_unref (newpad);
@@ -838,7 +954,8 @@ gst_rtp_bin_change_state (GstElement *element, GstStateChange transition)
             }
             if (!gst_rtp_bin_setup_elements (rtp_bin))
             {
-                GST_ERROR("Could not enter READY state because the elements could not be setup");
+                GST_ERROR("Could not enter READY state because the elements "
+                    "could not be setup");
                 goto start_failed;
             }
             break;
@@ -882,20 +999,22 @@ gst_rtp_bin_set_destinations (GstRTPBin *rtp_bin, const gchar *destination)
         if (g_strv_length (split_ip) != 2)
         {
             GST_DEBUG("Badly formated address %s, use ip:port format", ips[i]);
+            g_strfreev (split_ip);
             continue;
         }
 
 #ifdef HAVE_JRTP
         jrtpsession_setdestinationaddrs (rtp_bin->sess, split_ip[0],
                 atoi(split_ip[1]));
+        g_strfreev (split_ip);
 #else
         GST_DEBUG("Jrtplib not enabled, taking first destination only %s",
                 ips[i]);
         rtp_bin->dest_ip = inet_addr(split_ip[0]);
         rtp_bin->dest_port = htons(atoi(split_ip[1]));
+        g_strfreev (split_ip);
         break;
 #endif
-        g_strfreev (split_ip);
     }
     g_strfreev (ips);
 }
@@ -910,7 +1029,7 @@ gst_rtp_bin_set_property (GObject * object, guint prop_id,
 
     rtp_bin = GST_RTP_BIN (object);
 
-    GST_DEBUG("Setting props");
+    GST_DEBUG("Setting props: %d", prop_id);
 
     switch (prop_id) {
         case ARG_RTCP_SUPPORT:
@@ -949,14 +1068,22 @@ gst_rtp_bin_set_property (GObject * object, guint prop_id,
         case ARG_RTP_SOCKFD:
             rtp_bin->rtp_sockfd = g_value_get_int (value);
             if (rtp_bin->rtpsink != NULL)
+            {
               g_object_set (G_OBJECT(rtp_bin->rtpsink), "sockfd",
                   rtp_bin->rtp_sockfd, NULL);
+              g_object_set (G_OBJECT(rtp_bin->rtpsink), "closefd",
+                  rtp_bin->closefd, NULL);
+            }
             break;
         case ARG_RTCP_SOCKFD:
             rtp_bin->rtcp_sockfd = g_value_get_int (value);
             if (rtp_bin->rtcpsink != NULL)
+            {
               g_object_set (G_OBJECT(rtp_bin->rtcpsink), "sockfd",
                   rtp_bin->rtcp_sockfd, NULL);
+              g_object_set (G_OBJECT(rtp_bin->rtcpsink), "closefd",
+                  rtp_bin->closefd, NULL);
+            }
             break;
         case ARG_PT_MAP:
             rtp_bin->user_pt_map = (GHashTable *) g_value_get_pointer (value);
@@ -988,6 +1115,9 @@ gst_rtp_bin_set_property (GObject * object, guint prop_id,
               g_object_set (rtp_bin->jbuf, "queue_delay", rtp_bin->queue_delay,
                   NULL);
             break;
+        case ARG_CLOSE_FD:
+            rtp_bin->closefd = g_value_get_boolean (value);
+            break;
         default:
             G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
             break;
@@ -1035,6 +1165,10 @@ gst_rtp_bin_get_property (GObject * object, guint prop_id, GValue * value,
         case ARG_QUEUE_DELAY:
 	    g_value_set_uint (value, rtp_bin->queue_delay);
 	    break;
+        case ARG_CLOSE_FD:
+	    g_value_set_boolean (value, rtp_bin->closefd);
+	    break;
+
         default:
             G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
             break;
